----------------------------------------------------------------------
--- Operations to read the documentation comments in a Curry program.
---
--- @author Michael Hanus
----------------------------------------------------------------------

module AbstractCurryComments(readCurryWithComments,readComments) where

import Char
import AbstractCurry
import Directory(doesFileExist)

--------------------------------------------------------------------------
--- I/O action which parses a Curry program and returns the corresponding
--- typed Abstract Curry program. In addition to the operation
--- <code>AbstractCurry.readCurry</code>, this I/O action also reads
--- the documentation comments in the source file and puts
--- the function comments into the function declarations of
--- the Abstract Curry program (i.e., it uses the constructor
--- <code>CmtFunc</code> instead of <code>CFunc</code> whenever possible).

readCurryWithComments :: String -> IO CurryProg
readCurryWithComments progname = do
  prog <- readCurry progname
  let sourceFileName = progname++".curry"
  existsCurry <- doesFileExist sourceFileName
  if not existsCurry
   then return prog
   else do (_,_,fcmts) <- readComments sourceFileName
           return (addCommentsInProg fcmts prog)
 where
  addCommentsInProg fcmts (CurryProg mname imps types fdecls ops) =
    CurryProg mname imps types (map (addCommentsInFunc fcmts) fdecls) ops

  addCommentsInFunc fcmts fdecl@(CFunc fname ar vis ftype rules) =
    let fcmt = fcmts (snd fname)
     in if null fcmt
        then fdecl
        else CmtFunc fcmt fname ar vis ftype rules
  addCommentsInFunc _ fdecl@(CmtFunc _ _ _ _ _ _) = fdecl

--------------------------------------------------------------------------
--- Reads all documentation comments of a source file.
--- This operation returns the module comment,
--- a mapping from type names into their documentation comments,
--- and a mapping from function names into their documentation comments.
readComments :: String -> IO (String,String->String,String->String)
readComments filename = do
  (modcmts,cmtlines) <- readAllComments filename
  return (modcmts,
          \tname -> getDataComment tname cmtlines,
          \fname -> getFuncComment fname cmtlines)
fst3 (x,_,_) = x
trd3 (_,_,x) = x
main f = do r <- readComments "SetFunctions.curry"
            putStrLn (trd3 r f)

-- Reads all documentation comments of a source file.
readAllComments :: String -> IO (String,[(SourceLine,String)])
readAllComments filename = do
  prog <- readFile filename
  return (groupLines . filter (/=OtherLine) . map classifyLine . lines $ prog)

--- This datatype is used to classify all input lines.
--- @cons Comment   - a comment for CurryDoc
--- @cons FuncDef   - a definition of a function
--- @cons DataDef   - a definition of a datatype
--- @cons ModDef    - a line containing a module definition
--- @cons OtherLine - a line not relevant for CurryDoc
data SourceLine = Comment String  -- a comment for CurryDoc
                | FuncDef String  -- a definition of a function
                | DataDef String  -- a definition of a datatype
                | ModDef          -- a line containing a module definition
                | OtherLine       -- a line not relevant for CurryDoc

-- classify a line of the source program:
-- here we replace blank line comments by a "breakline" tag
classifyLine :: String -> SourceLine
classifyLine line
 | take 3 line == "---" && all isSpace (drop 3 line) = Comment "<br/>"
 | take 4 line == "--- "     = Comment (drop 4 line)
 | take 7 line == "module "  = ModDef
 | take 7 line == "import "  = ModDef
 | otherwise = let id1 = getFirstId line
                in if id1==""
                    then OtherLine
                    else if id1=="data" || id1=="type"
                          then DataDef (getDatatypeName line)
                          else FuncDef id1
 where
   getDatatypeName = takeWhile isIdChar . dropWhile (==' ') . dropWhile isIdChar

-- get the first identifier (name or operator in brackets) in a string:
getFirstId [] = ""
getFirstId (c:cs) | isAlpha c = takeWhile isIdChar (c:cs)
                  | c=='('    = takeWhile (/=')') cs
                  | otherwise = ""

-- is an alphanumeric character, underscore, or apostroph?
isIdChar c = isAlphaNum c || c=='_' || c=='\''


-- group the classified lines into module comment and list of
-- (Func/DataDef,comment) pairs:
groupLines :: [SourceLine] -> (String,[(SourceLine,String)])
groupLines sls =
  let (modcmts,progcmts) = break (==ModDef) sls
   in if progcmts==[]
      then ("", groupProgLines sls)
      else (concatMap getComment modcmts,
            groupProgLines (filter (/=ModDef) (tail progcmts)))
 where
   getComment (Comment cmt) = cmt++"\n"
   getComment (FuncDef _)   = ""  -- this case should usually not occur
   getComment (DataDef _)   = ""  -- this case should usually not occur

groupProgLines :: [SourceLine] -> [(SourceLine,String)]
groupProgLines [] = []
groupProgLines (Comment cmt : sls) = groupComment cmt sls
groupProgLines (FuncDef f : sls) = (FuncDef f, "") : skipFuncDefs f sls
groupProgLines (DataDef d : sls) = (DataDef d, "") : skipDataDefs d sls

groupComment _ [] = []  -- comment not followed by definition -> ignore
groupComment cmt (Comment cmt1 : sls) = groupComment (cmt++"\n"++cmt1) sls
groupComment cmt (FuncDef f : sls) = (FuncDef f, cmt) : skipFuncDefs f sls
groupComment cmt (DataDef d : sls) = (DataDef d, cmt) : skipDataDefs d sls

skipFuncDefs _ [] = []
skipFuncDefs _ (Comment cmt : sls) = groupProgLines (Comment cmt : sls)
skipFuncDefs _ (DataDef d   : sls) = groupProgLines (DataDef d   : sls)
skipFuncDefs f (FuncDef f1  : sls) =
  if f==f1 then skipFuncDefs f sls
           else groupProgLines (FuncDef f1 : sls)

skipDataDefs _ [] = []
skipDataDefs _ (Comment cmt : sls) = groupProgLines (Comment cmt : sls)
skipDataDefs _ (FuncDef f   : sls) = groupProgLines (FuncDef f   : sls)
skipDataDefs d (DataDef d1  : sls) =
  if d==d1 then skipDataDefs d sls
           else groupProgLines (DataDef d1 : sls)


--------------------------------------------------------------------------
-- get comment for a function name:
getFuncComment :: String -> [(SourceLine,String)] -> String
getFuncComment _ [] = ""
getFuncComment fname ((FuncDef f, cmt):fdcmts) =
  if fname == f
  then cmt
  else getFuncComment fname fdcmts
getFuncComment fname ((DataDef _,_):fdcmts) = getFuncComment fname fdcmts

-- get comment for a type name:
getDataComment :: String -> [(SourceLine,String)] -> String
getDataComment _ [] = ""
getDataComment n ((DataDef d, cmt):fdcmts) =
  if n == d then cmt
            else getDataComment n fdcmts 
getDataComment n ((FuncDef _,_):fdcmts) = getDataComment n fdcmts


--------------------------------------------------------------------------
