------------------------------------------------------------------------
-- Auxiliaries for assertion checking w.r.t. pre/postconditions.
------------------------------------------------------------------------

module Assert(Assert,makeAssert,waitOf,ddunifOf,
              aInt,aFloat,aChar,aBool,aOrdering,aSuccess,
              aList,aString,aPair,aTriple,aTuple4,aMaybe,aEither,
              aPolyType,aFun,
              enforce, enforceIO, enforceAssertions,
              checkPre, checkPost,
              withEnforceContract1,
              withEnforceContract2,
              withEnforcePostContract1,
              withEnforcePostContract2,
              withLazyContract1,
              withLazyContract2,
              withLazyPostContract1,
              withLazyPostContract2
             )  where

--- Auxiliaries for pre/postcondition checking

import SetFunctions
import GlobalVariable
import System(getEnviron)
import Unsafe

-- Should we show additional information about assertions?
--debug = False
debug = unsafePerformIO getDebugVar

getDebugVar = do
  v <- getEnviron "ASSERTDEBUG"
  return (v=="yes")

---------------------------------------------------------------------------
--- The main operation for enforceable assertions.
--- (enfore e) evaluates the expression e (with assertions)
--- so that unevaluated assertions are enforced before returning the result.
enforce :: a -> a
enforce exp | (unsafePerformIO initAssertionQueue =:= ())
              &> exp=:=exp
            = (unsafePerformIO activateAssertions =:= ())
              &> exp

--- Evaluate an I/O action with enforceable assertions.
--- (enforeIO a) evaluates the I/O action a (with assertions)
--- so that unevaluated assertions are enforced after performing action a.
enforceIO :: IO () -> IO ()
enforceIO a = initAssertionQueue >> a >> activateAssertions

-- Enforce immediate checking of all (delayed) assertions.
enforceAssertions :: IO ()
enforceAssertions = activateAssertions

---------------------------------------------------------------------------
-- Report the result of checking the pre/postconditions.
-- The first argument is a tag (the name of the operation).
-- The second argument is unified with () (used by enforceable constraints).
-- The third argument is a set of values that must not be empty for
-- a satisfied condition.
-- The fourth argument is a string describing the context (arguments,result).

checkPre :: String -> () -> Values _ -> String -> Success
checkPre fname eflag valset arg =
  if isEmpty valset
  then eflag=:=() &>
       error ("Precondition of operation '"++fname++"' failed for:\n"++arg)
  else eflag=:=() &> 
       if debug
       then unsafePerformIO
              (putStrLn ("Precondition of '"++fname++"' satisfied for: "++arg))
             =:= ()
       else success

checkPost :: String -> () -> Values _ -> String -> Success
checkPost fname eflag valset arg =
  if isEmpty valset
  then eflag=:=() &>
       error ("Postcondition of operation '"++fname++"' failed for:\n"++arg)
  else eflag=:=() &> 
       if debug
       then unsafePerformIO
              (putStrLn ("Postcondition of '"++fname++"' satisfied for: "++arg))
             =:= ()
       else success

---------------------------------------------------------------------------
-- Combinators for lazy checking of contracts with preconditions:

withLazyContract1 :: Assert a -> Assert b -> String
                  -> (a -> Values _) -> (a -> b -> Values _)
                  -> (a -> b) -> a -> b
withLazyContract1 (Assert waita ddunifa) (Assert waitb ddunifb)
                  fname presetfun postsetfun fun arg =
  spawnConstraint (checkPre fname () (presetfun (waita argx)) (show argx)
                   &
                   checkPost fname () (postsetfun (waita argx) (waitb fx))
                             (unwords [show argx,"->",show fx]))
                  (ddunifb fx (fun (ddunifa argx arg)))
 where
  argx, fx free

withLazyContract2 :: Assert a -> Assert b -> Assert c -> String
                  -> (a -> b -> Values _) -> (a -> b -> c -> Values _)
                  -> (a -> b -> c) -> a -> b -> c
withLazyContract2 (Assert waita ddunifa) (Assert waitb ddunifb)
                  (Assert waitc ddunifc)
                    fname presetfun postsetfun fun arg1 arg2 =
  spawnConstraint (checkPre fname ()
                        (presetfun (waita x1) (waitb x2))
                        (unwords [show x1,show x2])
                   &
                   checkPost fname ()
                        (postsetfun (waita x1) (waitb x2) (waitc fx))
                        (unwords [show x1,show x2,"->",show fx]))
                  (ddunifc fx (fun (ddunifa x1 arg1) (ddunifb x2 arg2)))
 where
  x1,x2,fx free

---------------------------------------------------------------------------
-- Combinators for lazy checking of contracts without preconditions:

withLazyPostContract1 :: Assert a -> Assert b -> String
                      -> (a -> b -> Values _)
                      -> (a -> b) -> a -> b
withLazyPostContract1 (Assert waita ddunifa) (Assert waitb ddunifb)
                      fname postsetfun fun arg =
  spawnConstraint (checkPost fname () (postsetfun (waita argx) (waitb fx))
                             (unwords [show argx,"->",show fx]))
                  (ddunifb fx (fun (ddunifa argx arg)))
 where
  argx,fx free

withLazyPostContract2 :: Assert a -> Assert b -> Assert c -> String
                      -> (a -> b -> c -> Values _)
                      -> (a -> b -> c) -> a -> b -> c
withLazyPostContract2 (Assert waita ddunifa) (Assert waitb ddunifb)
                      (Assert waitc ddunifc)
                      fname postsetfun fun arg1 arg2 =
  spawnConstraint (checkPost fname ()
                        (postsetfun (waita x1) (waitb x2) (waitc fx))
                        (unwords [show x1,show x2,"->",show fx]))
                  (ddunifc fx (fun (ddunifa x1 arg1) (ddunifb x2 arg2)))
 where
  x1,x2,fx free

---------------------------------------------------------------------------
-- Combinators for enforceable checking of contracts with preconditions:

withEnforceContract1 :: Assert a -> Assert b -> String
                     -> (a -> Values _) -> (a -> b -> Values _)
                     -> (a -> b) -> a -> b
withEnforceContract1 (Assert waita ddunifa) (Assert waitb ddunifb)
                     fname presetfun postsetfun fun arg
  | registerAssertion ef1 fname "Pre"
                      (checkPre fname () (presetfun arg) (show arg))
  & registerAssertion ef2 fname "Post"
      (let res = fun arg
        in checkPost fname () (postsetfun arg res)
                     (unwords [show arg,"->",show res]))
  = spawnConstraint (checkPre fname ef1 (presetfun (waita argx)) (show argx)
                     & 
                     checkPost fname ef2 (postsetfun (waita argx) (waitb fx))
                               (unwords [show argx,"->",show fx]))
                    (ddunifb fx (fun (ddunifa argx arg)))
 where
  argx,fx,ef1,ef2 free

withEnforceContract2 :: Assert a -> Assert b -> Assert c -> String
                     -> (a -> b -> Values _) -> (a -> b -> c -> Values _)
                     -> (a -> b -> c) -> a -> b -> c
withEnforceContract2 (Assert waita ddunifa) (Assert waitb ddunifb)
                     (Assert waitc ddunifc)
                     fname presetfun postsetfun fun arg1 arg2
  | registerAssertion ef1 fname "Pre"
                      (checkPre fname () (presetfun arg1 arg2)
                                (unwords [show arg1,show arg2]))
  & registerAssertion ef2 fname "Post"
      (let res = fun arg1 arg2
        in checkPost fname () (postsetfun arg1 arg2 res)
                     (unwords [show arg1,show arg2,"->",show res]))
  = spawnConstraint (checkPre fname ef1 (presetfun (waita x1) (waitb x2))
                              (unwords [show x1,show x2])
                     & 
                     checkPost fname ef2
                               (postsetfun (waita x1) (waitb x2) (waitc fx))
                               (unwords [show x1,show x2,"->",show fx]))
                    (ddunifc fx (fun (ddunifa x1 arg1) (ddunifb x2 arg2)))
 where
  x1,x2,fx,ef1,ef2 free

---------------------------------------------------------------------------
-- Combinators for enforceable checking of contracts without preconditions:

withEnforcePostContract1 :: Assert a -> Assert b -> String
                         -> (a -> b -> Values _)
                         -> (a -> b) -> a -> b
withEnforcePostContract1 (Assert waita ddunifa) (Assert waitb ddunifb)
                         fname postsetfun fun arg
  | registerAssertion ef fname "Post"
      (let res = fun arg
        in checkPost fname () (postsetfun arg res)
                      (unwords [show arg,"->",show res]))
  = spawnConstraint (checkPost fname ef (postsetfun (waita argx) (waitb fx))
                               (unwords [show argx,"->",show fx]))
                    (ddunifb fx (fun (ddunifa argx arg)))
 where
  argx,fx,ef free

withEnforcePostContract2 :: Assert a -> Assert b -> Assert c -> String
                         -> (a -> b -> c -> Values _)
                         -> (a -> b -> c) -> a -> b -> c
withEnforcePostContract2 (Assert waita ddunifa) (Assert waitb ddunifb)
                         (Assert waitc ddunifc)
                         fname postsetfun fun arg1 arg2
  | registerAssertion ef fname "Post"
      (let res = fun arg1 arg2
        in checkPost fname () (postsetfun arg1 arg2 res)
                     (unwords [show arg1,show arg2,"->",show res]))
  = spawnConstraint (checkPost fname ef
                        (postsetfun (waita x1) (waitb x2) (waitc fx))
                        (unwords [show x1,show x2,"->",show fx]))
                    (ddunifc fx (fun (ddunifa x1 arg1) (ddunifb x2 arg2)))
 where
  x1,x2,fx,ef free

---------------------------------------------------------------------------
-- Definition of the generic type for specification assertions on variaous
-- concrete types.

data Assert a = Assert (a->a) (a->a->a)

--- Constructor for Assert instances.
--- It creates from a wait and a demand-driven unification operation for
--- a given type an Assert instance for this type by generic handler
--- for logic variables as expression in the demand-driven unification.
makeAssert :: (a->a) -> (a->a->a) -> Assert a
makeAssert wait ddunif = Assert wait (withVarCheck ddunif)

--- Returns the wait operation of an Assert instance.
waitOf :: Assert a -> (a->a)
waitOf (Assert wait _) = wait

--- Returns the ddunif operation of an Assert instance.
ddunifOf :: Assert a -> (a->a->a)
ddunifOf (Assert _ ddunif) = ddunif

-- Auxiliary operation for demand-driven unification.
-- If the third argument is a logic variable, unify it with the second
-- (which is always a logic variable for passing the result) and return it,
-- otherwise apply the first argument (the "real" demand-driven unification).
withVarCheck :: (a -> a -> a) -> a -> a -> a
withVarCheck ddunif x e = if isVar e then (x=:=e) &> e else ddunif x e

-- Assert instances for various concrete types:

--- Assert instance for integers.
aInt :: Assert Int
aInt = aPrimType

--- Assert instance for floats.
aFloat :: Assert Float
aFloat = aPrimType

--- Assert instance for characters.
aChar :: Assert Char
aChar = aPrimType

--- Assert instance for Booleans.
aBool :: Assert Bool
aBool = aPrimType

--- Assert instance for Ordering.
aOrdering :: Assert Ordering
aOrdering = aPrimType

--- Assert instance for Success.
aSuccess :: Assert Success
aSuccess = aPrimType

--- Assert instance scheme for primitive (unstructured) types.
aPrimType :: Assert a
aPrimType = Assert ensureNotFree ddunifPrimType
  where
   ddunifPrimType x i | x =:= i = i

--- Assert instance scheme for polymorphic values.
--- Since we cannot examine the concrete structure of polymorphic values,
--- we simply wait until the complete value becomes evaluated to a ground term.
aPolyType :: Assert a
aPolyType = Assert ensureGround ddunifPolyType
  where
   ensureGround x = ensureNotFree (id $## x)
   ddunifPolyType x i | x =:= i = i

--- Assert instance for functional types.
--- Such assertions are implemented as assertions for primitive types
--- since functional objects are treated as constructors.
aFun :: Assert a -> Assert b -> Assert (a -> b)
aFun _ _ = aPrimType

--- Assert instance for lists.
aList :: Assert a -> Assert [a]
aList (Assert waita ddunifa) = makeAssert waitList ddunifList
 where
  waitList l = case l of []     -> []
                         (x:xs) -> waita x : waitList xs

  ddunifList x e = --if isVar e then (x=:=e) &> e else
    case e of
      []   -> (x=:=[])   &> []
      y:ys -> let z,zs free
               in (x=:=z:zs) &> (ddunifa z y : ddunifList zs ys)

--- Assert instance for string.
aString :: Assert String
aString = aList aChar

--- Assert instance for pairs.
aPair :: Assert a -> Assert b -> Assert (a,b)
aPair (Assert waita ddunifa) (Assert waitb ddunifb) =
  makeAssert waitPair ddunifPair
 where
  waitPair l = case l of (a,b) -> (waita a, waitb b)

  ddunifPair x e =
    case e of
      (a1,a2) -> let x1,x2 free
                  in x=:=(x1,x2) &> (ddunifa x1 a1, ddunifb x2 a2)

--- Assert instance for triples.
aTriple :: Assert a -> Assert b -> Assert c -> Assert (a,b,c)
aTriple (Assert waita ddunifa) (Assert waitb ddunifb) (Assert waitc ddunifc) =
  makeAssert waitTriple ddunifTriple
 where
  waitTriple l = case l of (a,b,c) -> (waita a, waitb b, waitc c)

  ddunifTriple x e =
    case e of
      (a1,a2,a3) -> let x1,x2,x3 free
                     in x=:=(x1,x2,x3) &>
                        (ddunifa x1 a1, ddunifb x2 a2, ddunifc x3 a3)

--- Assert instance for 4-tuples.
aTuple4 :: Assert a -> Assert b -> Assert c -> Assert d -> Assert (a,b,c,d)
aTuple4 (Assert waita ddunifa) (Assert waitb ddunifb) (Assert waitc ddunifc)
        (Assert waitd ddunifd) =
  makeAssert waitTuple ddunifTuple
 where
  waitTuple l = case l of (a,b,c,d) -> (waita a, waitb b, waitc c, waitd d)

  ddunifTuple x e =
    case e of
      (a1,a2,a3,a4) ->
         let x1,x2,x3,x4 free
          in x=:=(x1,x2,x3,x4) &>
             (ddunifa x1 a1, ddunifb x2 a2, ddunifc x3 a3, ddunifd x4 a4)

--- Assert instance for Maybe values.
aMaybe :: Assert a -> Assert (Maybe a)
aMaybe (Assert waita ddunifa) = makeAssert waitMaybe ddunifMaybe
 where
  waitMaybe e = case e of Nothing  -> Nothing
                          (Just x) -> Just (waita x)

  ddunifMaybe x e =
    case e of
      Nothing -> (x=:=Nothing) &> Nothing
      (Just y) ->  let z free
                    in (x =:= Just z) &> Just (ddunifa z y)

--- Assert instance for Either values.
aEither :: Assert a -> Assert b -> Assert (Either a b)
aEither (Assert waita ddunifa) (Assert waitb ddunifb) =
   makeAssert waitEither ddunifEither
 where
  waitEither e = case e of (Left  x) -> Left  (waita x)
                           (Right x) -> Right (waitb x)

  ddunifEither x e =
    case e of
      (Left  y) ->  let z free
                    in (x =:= Left  z) &> Left  (ddunifa z y)
      (Right y) ->  let z free
                    in (x =:= Right z) &> Right (ddunifb z y)


------------------------------------------------------------------------
-- Implement delaying and activating of assertion via global variables:

-- The global variable for assertion control:
assertionControl :: GVar ()
assertionControl = gvar unknown

-- Initialize assertion control:
initAssertionQueue :: IO ()
initAssertionQueue = do
  --putStrLn "INIT ASSERTION QUEUE"
  writeGVar assertionControl unknown

-- Get control variable for assertion. It is a logic variable
-- until the strict assertion checking is activated.
getAssertionControlVariable :: IO ()
getAssertionControlVariable = do
  v <- readGVar assertionControl
  --putStrLn ("Contol variable: "++showAnyExpression v)
  return v

-- Activate strict assertion checking by instantiation of control variable:
activateAssertions :: IO ()
activateAssertions = do
  --putStrLn "ACTIVATE ASSERTIONS"
  v <- getAssertionControlVariable
  doSolve (v=:=())
  initAssertionQueue -- new assertion control

------------------
-- Registers a new assertion for activation at some later point.
-- If the first argument is instantiated at the time of activation,
-- the assertion will be ignored, otherwise it will be evaluated.
registerAssertion :: () -> String -> String -> Success -> Success
registerAssertion eflag fname pp asrt
  | -- in order to be sure that the global control variable is initialized
    -- in the spawned constraint:
    unsafePerformIO (getAssertionControlVariable >> done) =:= ()
  = spawnConstraint (delayedAssertion eflag fname pp asrt =:= ())
                    success

delayedAssertion :: () -> String -> String -> Success -> ()
delayedAssertion eflag fname pp asrt = unsafePerformIO $ do
  v <- getAssertionControlVariable
  --putStrLn ("Delay assertion '"++fname++"': "++showAnyExpression asrt)
  ensureNotFree v =:= ()  &> done
  --putStrLn ("Assertion '"++fname++"' activated: "++showAnyExpression asrt)
  if isVar eflag
   then do if debug
            then putStrLn (pp++"condition of '"++fname++"' enforced...")
            else done
           asrt &> done
   else done

------------------------------------------------------------------------
