------------------------------------------------------------------------
--- A transformation from a Curry program with pre/postconditions
--- into a Curry program where these conditions are integrated
--- into the code
------------------------------------------------------------------------

import AbstractCurry
import AbstractCurryComments
import PrettyAbstract
import AbstractCurryGoodies
import System
import FileGoodies(stripSuffix)
import List
import Directory
import Time
import Distribution

banner = bannerLine ++ '\n' : bannerText ++ '\n' : bannerLine
 where
   bannerText = "DSDCurry Transformation Tool (Version of 16/07/10)"
   bannerLine = take (length bannerText) (repeat '=')

------------------------------------------------------------------------
-- Data type for transformation parameters
data TParam = TParam Bool -- generate code for lazy assertions?
                     Bool -- should I use enforcable lazy assertions?
                     Bool -- load and execute transformed program?
                     Bool -- debug assertions during execution?

defaultTParam = TParam False False False False

setStrictAssert  (TParam _ _ ep db) = TParam False False ep db

isLazyAssert     (TParam la _ _ _) = la

setLazyAssert    (TParam _ _ ep db) = TParam True False ep db

isEnforceAssert  (TParam _ fa _ _) = fa

setEnforceAssert (TParam _ _ ep db) = TParam True True ep db

executeProg      (TParam _ _ execprog _) = execprog

setExecuteProg   (TParam la fa _ db) = TParam la fa True db

withDebugging    (TParam _ _ _ db) = db

setDebugging     (TParam la fa ep _) = TParam la fa ep True

------------------------------------------------------------------------

main installdir = do
  putStrLn banner
  args <- getArgs
  processArgs defaultTParam args
 where
  processArgs tparam args = case args of
     ("-s":moreargs) -> processArgs (setStrictAssert  tparam) moreargs
     ("-l":moreargs) -> processArgs (setLazyAssert    tparam) moreargs
     ("-f":moreargs) -> processArgs (setEnforceAssert tparam) moreargs
     ("-r":moreargs) -> processArgs (setExecuteProg   tparam) moreargs
     ("-d":moreargs) -> processArgs (setDebugging     tparam) moreargs
     [mname]         -> transform installdir tparam (stripSuffix mname)
     _ -> putStrLn $ "ERROR: Illegal arguments for transformation: " ++
                     unwords args ++ "\n" ++
                     "Usage: dsdcurry [-s|-f|-l|-r|-d] <module_name>\n"++
                     "-s : generate strict assertions (default)\n"++
                     "-l : generate lazy assertions\n"++
                     "-f : generate lazy enforceable assertions\n"++
                     "-r : load the transformed program into PAKCS\n"++
                     "-d : show assertion results during execution (with -r)\n"

-- Specifies how the name of the transformed module is built from the
-- name of the original module.
transformedModName :: String -> String
transformedModName m = m++"C"

-- start PAKCS and load a module:
loadPAKCS tparam m = do
  putStrLn $ "\nStarting PAKCS and loading module '"++m++"'..."
  if withDebugging tparam then setEnviron "ASSERTDEBUG" "yes" else done
  system $ installDir++"/bin/pakcs -l "++m
  done

-- The main transformation function.
transform :: String -> TParam -> String -> IO ()
transform installdir tparam modname = do
  prog <- readCurryWithComments modname
  let preconds  = getPreConditions prog
      postconds = getPostConditions prog
      postnames = map (dropPostCondName . snd . funcName) postconds
      postfuncs = filter (\fd -> snd (funcName fd) `elem` postnames)
                         (filter isRealFuncDecl (funcDecls prog))
      newfuns   = postnames \\ (map (snd . funcName) postfuncs)
      checkfuns = map (snd . funcName) postfuncs
      onlyprec  = map (dropPreCondName . snd . funcName) preconds \\ postnames
      saveprog  = transPostConditions tparam prog postconds
      savefile  = transformedModName modname ++ ".curry"
  if null onlyprec then done else
    error ("Operations with preconditions but without postconditions: "
           ++ unwords onlyprec)
  if null newfuns then done else
    putStrLn $ "Generating new definitions for operations: " ++ unwords newfuns
  if null checkfuns then done else
    putStrLn $ "Generating pre/post-checked operations: " ++ unwords checkfuns
  if null postconds
   then putStrLn $ "No postconditions found, no transformation required!"
   else do putStrLn $ "Writing transformed module to '" ++ savefile ++ "'..."
           writeFile savefile (showCProg saveprog)
           copyImportModules installdir
           if executeProg tparam
            then loadPAKCS tparam (transformedModName modname)
            else done
  
-- Is a function declaration a real implementation, i.e.,
-- is the body different from "unknown"?
isRealFuncDecl :: CFuncDecl -> Bool
isRealFuncDecl (CFunc _ _ _ _ rs) = not (isUnknown rs)
isRealFuncDecl (CmtFunc _ _ _ _ _ rs) = not (isUnknown rs)

isUnknown :: CRules -> Bool
isUnknown (CExternal _) = False
isUnknown (CRules _ rs) = length rs == 1 && isUnknownRule (head rs)
 where
  isUnknownRule (CRule _ rhss _) = case rhss of
     [(_,e)] -> e == CSymbol ("Prelude","unknown")
     _       -> False

-- copy imported modules if necessary:
copyImportModules installdir = do
  cdir <- getCurrentDirectory
  if cdir==installdir
   then done
   else mapIO_ (\m -> copyOrUpdateFile installdir cdir (m++".curry"))
               ["Assert"]

-- get the preconditions from a Curry module
getPreConditions :: CurryProg -> [CFuncDecl]
getPreConditions prog = filter (isPreCondName . snd . funcName)
                               (funcDecls prog)

-- get the postconditions from a Curry module
getPostConditions :: CurryProg -> [CFuncDecl]
getPostConditions prog = filter (isPstCondName . snd . funcName)
                                (funcDecls prog)

transPostConditions :: TParam -> CurryProg -> [CFuncDecl] -> CurryProg
transPostConditions tparam (CurryProg mname imps tdecls fdecls opdecls)
                    postdecls =
 let postnames = map (\fd -> dropPostCondName (snd (funcName fd))) postdecls
     otherfdecls = filter (\fd -> snd (funcName fd) `notElem` postnames) fdecls
     rename (mn,fn) = (if mn==mname then transformedModName mname else mn, fn)
  in CurryProg (transformedModName mname)
               (nub ("Assert":"SetFunctions":imps))
               (map (renameSymbolInTypeDecl rename) tdecls)
               (map (renameSymbolInFunc rename)
                    (map (transPostCondition tparam fdecls) postdecls
                     ++ otherfdecls
                     ++ map genAssertInstance tdecls))
               (map (renameOpDecl rename) opdecls)

transPostCondition :: TParam -> [CFuncDecl] -> CFuncDecl -> CFuncDecl
transPostCondition tparam allfuncs (CFunc qn ar vis texp rules) =
  transPostCondition tparam allfuncs (CmtFunc "" qn ar vis texp rules)
transPostCondition tparam allfuncs (CmtFunc cmt (m,f) ar vis texp _) =
 let fname     = dropPostCondName f
     deffuncs  = filter (\fd -> snd (funcName fd) == fname) allfuncs
     deftype   = typeOf (head deffuncs) -- type of implemented operation
     noimpl    = null (filter isRealFuncDecl deffuncs)
     orgfunc   = head deffuncs -- original implementation of operation
     orgcmt    = if noimpl then cmt
                           else let orgfuncmt = commentOf orgfunc
                                 in if null orgfuncmt then cmt else orgfuncmt
     asrttypes = map toPolymorphicAssertTypes (getArgResultTypes deftype)
     preconds  = filter (\fd -> snd (funcName fd) == fname++"'pre") allfuncs
     argvars   = map (\i -> (i,"x"++show i)) [1..(ar-1)]
     numargs   = length argvars
     resultvar = (ar,"y")
     precall   = applyF ("SetFunctions","set"++show numargs)
                        (constF (funcName (head preconds)) : map CVar argvars)
     precheck  = applyF (aMod "checkPre")
                        [string2ac fname, constF (pre "()"), precall,
                         applyF (pre "unwords")
                                [list2ac
                                   (map (\a -> applyF (pre "show") [CVar a])
                                        argvars)]]
     postcall  = applyF (m,f) (map CVar (argvars++[resultvar]))
     postsetcall = applyF ("SetFunctions","set"++show (numargs+1))
                          (constF (m,f) : map CVar (argvars++[resultvar]))
     postcheck = applyF (aMod "checkPost")
                        [string2ac fname, constF (pre "()"), postsetcall,
                         applyF (pre "unwords")
                           [list2ac
                             (map (\a -> applyF (pre "show") [CVar a]) argvars ++
                              [string2ac "->",
                               applyF (pre "show") [CVar resultvar]])]]
     asrtCall  = if null preconds
                 then applyF (aMod $
                              (if isEnforceAssert tparam
                               then "withEnforcePostContract"
                               else "withLazyPostContract") ++show numargs)
                        (asrttypes ++
                         [string2ac fname,
                          applyF ("SetFunctions","set"++show (numargs+1))
                                 [constF (m,f)],
                          constF (m,fname++"'")] ++
                         map CVar argvars)
                 else applyF (aMod $
                              (if isEnforceAssert tparam
                               then "withEnforceContract"
                               else "withLazyContract") ++show numargs)
                        (asrttypes ++
                         [string2ac fname,
                          applyF ("SetFunctions","set"++show numargs)
                                 [constF (funcName (head preconds))],
                          applyF ("SetFunctions","set"++show (numargs+1))
                                 [constF (m,f)],
                          constF (m,fname++"'")] ++
                         map CVar argvars)
     rename qf = if qf==(m,fname) then (m,fname++"'") else qf
  in deleteCmtIfEmpty $ cmtfunc
           orgcmt
           (transformedModName m,dropPostCondName f) (ar-1) vis
           (if null deffuncs then transPCType f texp else deftype)
           [if noimpl -- no implementation provided: use pre/postcond
            then CRule (map CPVar argvars)
                       [(if null preconds
                         then postcall
                         else applyF (pre "&>") [precheck,postcall],
                         CVar resultvar)]
                       [CLocalVar resultvar]
            else if isLazyAssert tparam 
                 then CRule (map CPVar argvars)
                            [(constF (pre "success"),asrtCall)]
                             (map (renameSymbolInLocal rename)
                                  [CLocalFunc (deleteCmt orgfunc)])
                 else CRule (map CPVar argvars)
                            [(if null preconds
                              then postcheck
                              else applyF (pre "&>") [precheck,postcheck],
                              CVar resultvar)]
                            (map (renameSymbolInLocal rename)
                                 [CLocalPat (CPVar resultvar)
                                            (applyF (m,fname) (map CVar argvars))
                                            [],
                                  CLocalFunc (deleteCmt orgfunc)])]

-- Get types of arguments and result
getArgResultTypes t = case t of
  CFuncType t1 t2 -> t1 : getArgResultTypes t2
  _               -> [t]

-- Transforms a monomorphic type expression into an assertion instance
-- expression of this type.
toPolymorphicAssertTypes :: CTypeExpr -> CExpr
toPolymorphicAssertTypes =
  toAssertTypes (\_ -> constF (aMod "aPolyType"))

-- Transforms a type expression into an assertion instance expression
-- of this type. The first argument specifies the mapping for
-- type variables.
toAssertTypes :: ((Int,String) -> CExpr) -> CTypeExpr -> CExpr
toAssertTypes mapvar (CTVar tv) = mapvar tv
toAssertTypes mapvar (CFuncType t1 t2) =
  applyF (aMod "aFun") (map (toAssertTypes mapvar) [t1,t2])

toAssertTypes mapvar (CTCons (tm,tc) targs) =
  applyF (if tm=="Prelude" then aMod (preludeTCons2Assert tc) else (tm,'a':tc))
         (map (toAssertTypes mapvar) targs)
 where
  preludeTCons2Assert c = case c of
    "Int"      -> "aInt"
    "Float"    -> "aFloat"
    "Char"     -> "aChar"
    "Bool"     -> "aBool"
    "Success"  -> "aSuccess"
    "Ordering" -> "aOrdering"
    "[]"       -> "aList"
    "String"   -> "aString"
    "Maybe"    -> "aMaybe"
    "Either"   -> "aEither"
    "(,)"      -> "aPair"
    "(,,)"     -> "aTriple"
    "(,,,)"    -> "aTuple4"
    _          -> error ("Lazy assertion of type '"++c++"' not yet supported!")

 
-- Transforms the type of a postcondition into a type for the
-- corresponding function.
transPCType f t = case t of
  CFuncType t1 (CTCons _ _) -> t1
  CFuncType t1 t2 -> CFuncType t1 (transPCType f t2)
  _ -> error ("Illegal type of postcondition \""++f++"\"!")


-- Is this the name of a precondition?
isPreCondName f = take 4 (reverse f) == "erp'"

-- Is this the name of a precondition?
isPstCondName f = take 5 (reverse f) == "tsop'"

dropPreCondName f = reverse (drop 4 (reverse f))

dropPostCondName f = reverse (drop 5 (reverse f))

-- An operation of the assert module:
aMod f = ("Assert",f)


------------------------------------------------------------------------
-- Generate code for Assert instances on user-defined types.

genAssertInstance :: CTypeDecl -> CFuncDecl
genAssertInstance (CTypeSyn (mn,tc) _ targs texp) =
  cfunc
   (mn,'a':tc) 1 Public
   (foldr (\tv t -> CTCons (aMod "Assert") [CTVar tv] ~> t)
          (CTCons (aMod "Assert") [CTCons (mn,tc) (map CTVar targs)])
          targs)
   [CRule (map (\ (i,s) -> CPVar (i,"assert_"++s)) targs)
          [noGuard (toLocalAssertType texp)]
          []]

genAssertInstance (CType (mn,tc) _ targs tcs) =
  cfunc
   (mn,'a':tc) 1 Public
   (foldr (\tv t -> CTCons (aMod "Assert") [CTVar tv] ~> t)
          (CTCons (aMod "Assert") [mtype])
          targs)
   [CRule
     (map (\ (i,s) -> CPVar (i,"assert_"++s)) targs)
     [noGuard (applyF (aMod "makeAssert") [CSymbol (mn,"wait"++tc),
                                       CSymbol (mn,"ddunif"++tc)])]
     [CLocalFunc $
        cfunc (mn,"wait"++tc) 1 Private
              (mtype ~> mtype)
              [CRule [CPVar wvar]
                     [noGuard (CCase (CVar wvar) (map tc2waitbranch tcs))]
                     []],
      CLocalFunc $
        cfunc (mn,"ddunif"++tc) 1 Private
              (mtype ~> mtype ~> mtype)
              [CRule [CPVar uvar1, CPVar uvar2]
                     [noGuard (CCase (CVar uvar2) (map tc2unifbranch tcs))]
                     []]
     ]]
 where
  mtype = CTCons (mn,tc) (map CTVar targs)
  wvar = (1,"wv")
  uvar1 = (1,"x")
  uvar2 = (2,"exp")

  tc2waitbranch (CCons (_,cons) n _ texps) =
    CBranch (CPComb (mn,cons) (map (\i -> CPVar (i,'x':show i)) [1..n]))
            (applyF (mn,cons)
               (map (\ (i,texp) -> texp2wait texp [CVar (i,'x':show i)])
                    (zip [1..n] texps)))

  tc2unifbranch (CCons (_,cons) n _ texps) =
    CBranch (CPComb (mn,cons) (map (\i -> CPVar (i,'x':show i)) [1..n]))
            (clet
               (map (\i -> CLocalVar (i+100,'y':show i)) [1..n])
               (applyF (pre "&>")
                       [applyF (pre "=:=")
                               [CVar uvar1,
                                applyF (mn,cons)
                                  (map (\i -> CVar (i+100,'y':show i)) [1..n])],
                        applyF (mn,cons)
                          (map (\ (i,texp) -> 
                                 texp2unif texp
                                  [CVar (i+100,'y':show i), CVar (i,'x':show i)])
                               (zip [1..n] texps))]))

  texp2wait (CTVar (i,tv)) args = applyF (aMod "waitOf")
                                         (CVar (i,"assert_"++tv) : args)
  texp2wait texp@(CFuncType _ _) args =
    applyF (aMod "waitOf") (toLocalAssertType texp : args)
  texp2wait texp@(CTCons (tm,tcons) _) args =
    if texp==mtype -- recursive type?
    then applyF (tm,"wait"++tcons) args
    else applyF (aMod "waitOf") (toLocalAssertType texp : args)

  texp2unif (CTVar (i,tv)) args = applyF (aMod "ddunifOf")
                                         (CVar (i,"assert_"++tv) : args)
  texp2unif texp@(CFuncType _ _) args =
    applyF (aMod "ddunifOf") (toLocalAssertType texp : args)
  texp2unif texp@(CTCons (tm,tcons) _) args =
    if texp==mtype -- recursive type?
    then applyF (tm,"ddunif"++tcons) args
    else applyF (aMod "ddunifOf") (toLocalAssertType texp : args)

-- Transforms a type expression into an assertion instance expression
-- of this type.
toLocalAssertType = toAssertTypes (\ (i,tv) -> CVar (i,"assert_"++tv))


------------------------------------------------------------------------
-- Auxiliary operations:

--- Copy a file from a directory into another directory.
--- Do not copy it if there is already file with the same time stamp and size.
copyOrUpdateFile :: String -> String -> String -> IO ()
copyOrUpdateFile fdir tdir fname = do
  let orgfname = fdir ++ '/' : fname
      newfname = tdir ++ '/' : fname
      copycmd  = putStrLn ("Copying auxiliary module: "++fname) >>
                 system ("cp "++orgfname++" "++newfname) >> done
  exfile <- doesFileExist newfname
  if exfile
   then do odate <- getModificationTime orgfname
           ndate <- getModificationTime newfname
           osize <- fileSize orgfname
           nsize <- fileSize newfname
           if compareClockTime ndate odate /= LT && osize == nsize
            then done
            else copycmd
   else copycmd

------------------------------------------------------------------------
